package ru.dbotthepony.kstarbound.world.entities.behavior

import org.apache.logging.log4j.LogManager
import org.classdump.luna.exec.CallPausedException
import org.classdump.luna.lib.CoroutineLib
import org.classdump.luna.runtime.Coroutine
import ru.dbotthepony.kstarbound.defs.actor.behavior.NodeParameter
import ru.dbotthepony.kstarbound.lua.userdata.BehaviorState

class DecoratorNode(val name: String, val parameters: Map<String, NodeParameter>, val child: AbstractBehaviorNode) : AbstractBehaviorNode() {
	private var coroutine: Coroutine? = null

	override fun run(delta: Double, state: BehaviorState): Status {
		var coroutine = coroutine

		if (coroutine == null) {
			val parameters = state.blackboard.parameters(parameters, this)
			val fn = state.functions[name] ?: throw RuntimeException("How? $name")

			try {
				coroutine = state.lua.call(CoroutineLib.create(), fn)[0] as Coroutine
				val result = state.lua.call(CoroutineLib.resume(), coroutine, parameters, state.blackboard, this)
				val status = result[0] as Boolean

				if (!status && result.size >= 2) {
					LOGGER.warn("Behavior DecoratorNode '$name' failed: ${result[1]}")
				}

				return if (status) Status.SUCCESS else Status.FAILURE
			} catch (err: CallPausedException) {
				LOGGER.error("Behavior DecoratorNode '$name' called blocking code, which initiated pause. This is not supported.")
				return Status.FAILURE
			}
		}

		// decorator runs its child on yield and is resumed with the child's status on success or failure
		var status = Status.RUNNING

		while (status == Status.RUNNING) {
			val childStatus = child.runAndReset(delta, state)

			if (childStatus == Status.SUCCESS || childStatus == Status.FAILURE) {
				try {
					val result = state.lua.call(CoroutineLib.resume(), coroutine, childStatus.asBoolean)
					val execStatus = result[0] as Boolean

					if (!execStatus && result.size >= 2) {
						LOGGER.warn("Behavior DecoratorNode '$name' failed: ${result[1]}")
					}

					status = if (execStatus) Status.SUCCESS else Status.FAILURE
				} catch (err: CallPausedException) {
					LOGGER.error("Behavior DecoratorNode '$name' called blocking code on children return, which initiated pause. This is not supported.")
					status = Status.FAILURE
				}
			} else {
				return Status.RUNNING
			}
		}

		return status
	}

	override fun reset() {
		coroutine = null
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
